/*
    libutf8filename -- transparently convert utf8 filenames into 8 bit ones
    Copyright (C) 2003  Egmont Koblinger <egmont@uhulinux.hu>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#define _GNU_SOURCE
#include <unistd.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <fcntl.h>
#include <utime.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <locale.h>
#include <langinfo.h>
#include <iconv.h>
#include <limits.h>

static struct dirent *(*next_readdir) (DIR *);
static struct dirent64 *(*next_readdir64) (DIR *);
static int (*next_readdir_r) (DIR *, struct dirent *, struct dirent **);
static int (*next_readdir64_r) (DIR *, struct dirent64 *, struct dirent64 **);
static char *(*next_getcwd) (char *, size_t);
static char *(*next_get_current_dir_name) (void);
static char *(*next_getcw) (char *);
static int (*next_readlink) (const char *, char *, size_t);
static int (*next_chdir) (const char *);
static int (*next_open) (const char *, int, ...);
static int (*next_open64) (const char *, int, ...);
static int (*next_creat) (const char *, mode_t);
static int (*next_creat64) (const char *, mode_t);
static DIR *(*next_opendir) (const char *);
static FILE *(*next_fopen) (const char *, const char *);
static FILE *(*next_fopen64) (const char *, const char *);
static FILE *(*next_freopen) (const char *, const char *, FILE *);
static FILE *(*next_freopen64) (const char *, const char *, FILE *);
static int (*next_access) (const char *, int);
static int (*next___xstat) (int, const char *, struct stat *);
static int (*next___xstat64) (int, const char *, struct stat64 *);
static int (*next___lxstat) (int, const char *, struct stat *);
static int (*next___lxstat64) (int, const char *, struct stat64 *);
static int (*next_statfs) (const char *, struct statfs *);
static int (*next_chmod) (const char *, mode_t);
static int (*next_chown) (const char *, uid_t, gid_t);
static int (*next_lchown) (const char *, uid_t, gid_t);
static int (*next_utime) (const char *, const struct utimbuf *);
static int (*next_utimes) (const char *, const struct timeval *);
static int (*next_lutimes) (const char *, const struct timeval *);
static int (*next_link) (const char *, const char *);
static int (*next_symlink) (const char *, const char *);
static int (*next_mkdir) (const char *, mode_t);
static int (*next_mknod) (const char *, mode_t, dev_t);
static int (*next_rename) (const char *, const char *);
static int (*next_unlink) (const char *);
static int (*next_rmdir) (const char *);
static int (*next_truncate) (const char *, off_t);

static iconv_t ic_local2utf;
static iconv_t ic_utf2local;

static char *prefix = NULL;
static int prefixlen = 0;

static int mangle = 0;

void sortdir_init (void) __attribute((constructor));
void sortdir_init (void)
{
	locale_t locale = NULL;
	char *charsetname;
	char *localename;
	char *p;

	next_readdir              = dlsym(RTLD_NEXT, "readdir");
	next_readdir64            = dlsym(RTLD_NEXT, "readdir64");
	next_readdir_r            = dlsym(RTLD_NEXT, "readdir_r");
	next_readdir64_r          = dlsym(RTLD_NEXT, "readdir64_r");
	next_getcwd               = dlsym(RTLD_NEXT, "getcwd");
	next_get_current_dir_name = dlsym(RTLD_NEXT, "get_current_dir_name"); 
	next_getcw                = dlsym(RTLD_NEXT, "getcw");
	next_readlink             = dlsym(RTLD_NEXT, "readlink");
	next_chdir                = dlsym(RTLD_NEXT, "chdir");
	next_open                 = dlsym(RTLD_NEXT, "open");
	next_open64               = dlsym(RTLD_NEXT, "open64");
	next_creat                = dlsym(RTLD_NEXT, "creat");
	next_creat64              = dlsym(RTLD_NEXT, "creat64");
	next_opendir              = dlsym(RTLD_NEXT, "opendir");
	next_fopen                = dlsym(RTLD_NEXT, "fopen");
	next_fopen64              = dlsym(RTLD_NEXT, "fopen64");
	next_freopen              = dlsym(RTLD_NEXT, "freopen");
	next_freopen64            = dlsym(RTLD_NEXT, "freopen64");
	next_access               = dlsym(RTLD_NEXT, "access");
	next___xstat              = dlsym(RTLD_NEXT, "__xstat");
	next___xstat64            = dlsym(RTLD_NEXT, "__xstat64");
	next___lxstat             = dlsym(RTLD_NEXT, "__lxstat");
	next___lxstat64           = dlsym(RTLD_NEXT, "__lxstat64");
	next_statfs               = dlsym(RTLD_NEXT, "statfs");
	next_chmod                = dlsym(RTLD_NEXT, "chmod");
	next_chown                = dlsym(RTLD_NEXT, "chown");
	next_lchown               = dlsym(RTLD_NEXT, "lchown");
	next_utime                = dlsym(RTLD_NEXT, "utime");
	next_utimes               = dlsym(RTLD_NEXT, "utimes");
	next_lutimes              = dlsym(RTLD_NEXT, "lutimes");
	next_link                 = dlsym(RTLD_NEXT, "link");
	next_symlink              = dlsym(RTLD_NEXT, "symlink");
	next_mknod                = dlsym(RTLD_NEXT, "mknod");
	next_mkdir                = dlsym(RTLD_NEXT, "mkdir");
	next_rename               = dlsym(RTLD_NEXT, "rename");
	next_unlink               = dlsym(RTLD_NEXT, "unlink");
	next_rmdir                = dlsym(RTLD_NEXT, "rmdir");
	next_truncate             = dlsym(RTLD_NEXT, "truncate");

	charsetname = getenv("UTF8FILENAME_CHARSET");
	if (charsetname == NULL) {
		localename = getenv("UTF8FILENAME_LOCALE");
		locale = newlocale(LC_ALL_MASK, localename != NULL ? localename : "", NULL);
		if (locale == NULL) return;
		charsetname = nl_langinfo_l(CODESET, locale);
	}
	if (charsetname == NULL || charsetname[0] == 0) return;
	ic_local2utf = iconv_open("UTF-8", charsetname);
	if (ic_local2utf == (iconv_t)(-1)) return;
	ic_utf2local = iconv_open(charsetname, "UTF-8");
	if (ic_utf2local == (iconv_t)(-1)) return;
	if (locale != NULL) freelocale(locale);

	p = getenv("UTF8FILENAME_PREFIX");
	if (p != NULL && p[0] != 0 && strchr(p, '/') == NULL) {
		prefix = strdup(p);
		prefixlen = strlen(prefix);
	}

	mangle = 1;
}

/*--------------------------------------------------------------------------*/

static int real2fake_name (const char *real, char *fake, int buflen)
{
	const char *real2 = real;
	char *fake2 = fake;
	size_t inbytesleft = strcspn(real, "/");
	size_t outbytesleft = buflen - 1;
	iconv(ic_utf2local, NULL, 0, NULL, 0);
	if (iconv(ic_utf2local, &real, &inbytesleft, &fake, &outbytesleft) == (size_t)(-1)) {
		if (prefix == NULL) return -1;
		errno = 0;
		real = real2;
		fake = fake2;
		inbytesleft = strcspn(real, "/");
		outbytesleft = buflen - 1;
		if (inbytesleft + prefixlen > outbytesleft) {
			errno = ENAMETOOLONG;
			return -1;
		}
		strncpy(fake, prefix, prefixlen);
		strncpy(fake + prefixlen, real, inbytesleft);
		fake[inbytesleft + prefixlen] = 0;
		return inbytesleft + prefixlen;
	}
	*fake = 0;
	return buflen - 1 - outbytesleft;
}

static int real2fake_path (const char *real, char *fake, int buflen)
{
	int i;
	while (*real && buflen > 0) {
		if (*real == '/') {
			real++;
			*fake++ = '/';
			buflen--;
			continue;
		}
		i = real2fake_name(real, fake, buflen);
		if (i < 0) return -1;
		real += strcspn(real, "/");
		fake += i;
		buflen -= i;
	}
	if (buflen <= 0) {
		errno = ENAMETOOLONG;
		return -1;
	}
	*fake = 0;
	return 0;
}

static int fake2real_name (const char *fake, char *real, int buflen)
{
	size_t inbytesleft = strcspn(fake, "/");
	size_t outbytesleft = buflen - 1;
	if (prefix != NULL && !strncmp(fake, prefix, prefixlen)) {
		if (inbytesleft - prefixlen > outbytesleft) {
			errno = ENAMETOOLONG;
			return -1;
		}
		strncpy(real, fake + prefixlen, inbytesleft - prefixlen);
		real[inbytesleft - prefixlen] = 0;
		return inbytesleft - prefixlen;
	}
	iconv(ic_local2utf, NULL, 0, NULL, 0);
	if (iconv(ic_local2utf, &fake, &inbytesleft, &real, &outbytesleft) == (size_t)(-1)) return -1;
	*real = 0;
	return buflen - 1 - outbytesleft;
}

static int fake2real_path (const char *fake, char *real, int buflen)
{
	int i;
	while (*fake && buflen > 0) {
		if (*fake == '/') {
			fake++;
			*real++ = '/';
			buflen--;
			continue;
		}
		i = fake2real_name(fake, real, buflen);
		if (i < 0) return -1;
		fake += strcspn(fake, "/");
		real += i;
		buflen -= i;
	}
	if (buflen <= 0) {
		errno = ENAMETOOLONG;
		return -1;
	}
	*real = 0;
	return 0;
}

/*--------------------------------------------------------------------------*/

struct dirent *readdir (DIR *dir)
{
	struct dirent *dirent;
	char fake[NAME_MAX + 1];
	if (!mangle) return next_readdir(dir);
	while (1) {
		dirent = next_readdir(dir);
		if (dirent == NULL) return NULL;
		if (real2fake_name(dirent->d_name, fake, sizeof(fake)) < 0) continue;
		strcpy(dirent->d_name, fake);
#ifdef _DIRENT_HAVE_D_NAMLEN
		dirent->d_namlen = strlen(dirent->d_name);
#endif
		return dirent;
	}
}

struct dirent64 *readdir64 (DIR *dir)
{
	struct dirent64 *dirent;
	char fake[NAME_MAX + 1];
	if (!mangle) return next_readdir64(dir);
	while (1) {
		dirent = next_readdir64(dir);
		if (dirent == NULL) return NULL;
		if (real2fake_name(dirent->d_name, fake, sizeof(fake)) < 0) continue;
		strcpy(dirent->d_name, fake);
#ifdef _DIRENT_HAVE_D_NAMLEN
		dirent->d_namlen = strlen(dirent->d_name);
#endif
		return dirent;
	}
}

int readdir_r (DIR *dir, struct dirent *dirent, struct dirent **resultp)
{
	int retval;
	char fake[NAME_MAX + 1];
	if (!mangle) return next_readdir_r(dir, dirent, resultp);
	while (1) {
		next_readdir_r(dir, dirent, resultp);
		if (resultp == NULL || retval < 0) return retval;
		if (real2fake_name(dirent->d_name, fake, sizeof(fake)) < 0) continue;
		strcpy(dirent->d_name, fake);
#ifdef _DIRENT_HAVE_D_NAMLEN
		dirent->d_namlen = strlen(dirent->d_name);
#endif
		return 0;
	}
}

int readdir64_r (DIR *dir, struct dirent64 *dirent, struct dirent64 **resultp)
{
	int retval;
	char fake[NAME_MAX + 1];
	if (!mangle) return next_readdir64_r(dir, dirent, resultp);
	while (1) {
		next_readdir64_r(dir, dirent, resultp);
		if (resultp == NULL || retval < 0) return retval;
		if (real2fake_name(dirent->d_name, fake, sizeof(fake)) < 0) continue;
		strcpy(dirent->d_name, fake);
#ifdef _DIRENT_HAVE_D_NAMLEN
		dirent->d_namlen = strlen(dirent->d_name);
#endif
		return 0;
	}
}

char *getcwd (char *buf, size_t size)
{
	char *c;
	int i;
	char real[PATH_MAX + 1];
	char fake[PATH_MAX + 1];
	if (!mangle) return next_getcwd(buf, size);
	c = next_getcwd(real, PATH_MAX);
	if (c != real) return c;
	if (real2fake_path(real, fake, sizeof(fake)) < 0) return NULL;
	i = strlen(fake);
	if (size && i > size - 1) {
		errno = ERANGE;
		return NULL;
	}
	if (size) {
		strcpy(buf, fake);
	} else {
		buf = strdup(fake);
	}
	return buf;
}

char *get_current_dir_name (void)
{
	char *buf;
	char *real;
	char fake[PATH_MAX + 1];
	if (!mangle) return next_get_current_dir_name();
	real = next_get_current_dir_name();
	if (real == NULL) return NULL;
	if (real2fake_path(real, fake, sizeof(fake)) < 0) return NULL;
	free(real);
	buf = strdup(fake);
	return buf;
}

char *getcw (char *buf)
{
	char *c;
	char real[PATH_MAX + 1];
	char fake[PATH_MAX + 1];
	if (!mangle) return next_getcw(buf);
	c = next_getcw(real);
	if (c != real) return c;
	if (real2fake_path(real, fake, sizeof(fake)) < 0) return NULL;
	strcpy(buf, fake);
	return buf;
}

int readlink (const char *fake1, char *buf, size_t bufsiz)
{
	int i;
	char real1[PATH_MAX + 1];
	char real2[PATH_MAX + 1];
	char fake2[PATH_MAX + 1];
	if (!mangle) return next_readlink(fake1, buf, bufsiz);
	if (fake2real_path(fake1, real1, sizeof(real1)) < 0) return -1;
	i = next_readlink(real1, real2, PATH_MAX);
	if (i < 0) return i;
	real2[i] = 0;
	if (real2fake_path(real2, fake2, sizeof(fake2)) < 0) return -1;
	i = strlen(fake2);
	if (i > bufsiz) i = bufsiz;
	memcpy(buf, fake2, i);
	return i;
}

int chdir (const char *fake)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_chdir(fake);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_chdir(real);
}

int open (const char *fake, int flags, ...)
{
	mode_t mode;
	char real[PATH_MAX + 1];
	if (!mangle) {
		if (flags & O_CREAT) {
			va_list ap;
			va_start(ap, flags);
			mode = va_arg(ap, mode_t);
			va_end(ap);
			return next_open(fake, flags, mode);
		} else {
			return next_open(fake, flags);
		}
	}
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	if (flags & O_CREAT) {
		va_list ap;
		va_start(ap, flags);
		mode = va_arg(ap, mode_t);
		va_end(ap);
		return next_open(real, flags, mode);
	} else {
		return next_open(real, flags);
	}
}

int open64 (const char *fake, int flags, ...)
{
	mode_t mode;
	char real[PATH_MAX + 1];
	if (!mangle) {
		if (flags & O_CREAT) {
			va_list ap;
			va_start(ap, flags);
			mode = va_arg(ap, mode_t);
			va_end(ap);
			return next_open64(fake, flags, mode);
		} else {
			return next_open64(fake, flags);
		}
	}
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	if (flags & O_CREAT) {
		va_list ap;
		va_start(ap, flags);
		mode = va_arg(ap, mode_t);
		va_end(ap);
		return next_open64(real, flags, mode);
	} else {
		return next_open64(real, flags);
	}
}

int creat (const char *fake, mode_t mode)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_creat(fake, mode);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_creat(real, mode);
}

int creat64 (const char *fake, mode_t mode)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_creat64(fake, mode);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_creat64(real, mode);
}

DIR *opendir (const char *fake)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_opendir(fake);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return NULL;
	return next_opendir(real);
}

FILE *fopen (const char *fake, const char *mode)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_fopen(fake, mode);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return NULL;
	return next_fopen(real, mode);
}

FILE *fopen64 (const char *fake, const char *mode)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_fopen64(fake, mode);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return NULL;
	return next_fopen64(real, mode);
}

FILE *freopen (const char *fake, const char *mode, FILE *f)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_freopen(fake, mode, f);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return NULL;
	return next_freopen(real, mode, f);
}

FILE *freopen64 (const char *fake, const char *mode, FILE *f)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_freopen64(fake, mode, f);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return NULL;
	return next_freopen64(real, mode, f);
}

int access (const char *fake, int mode)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_access(fake, mode);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_access(real, mode);
}

int __xstat (int ver, const char *fake, struct stat *st)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next___xstat(ver, fake, st);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next___xstat(ver, real, st);
}

int __xstat64 (int ver, const char *fake, struct stat64 *st)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next___xstat64(ver, fake, st);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next___xstat64(ver, real, st);
}

int __lxstat (int ver, const char *fake, struct stat *st)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next___lxstat(ver, fake, st);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next___lxstat(ver, real, st);
}

int __lxstat64 (int ver, const char *fake, struct stat64 *st)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next___lxstat64(ver, fake, st);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next___lxstat64(ver, real, st);
}

int statfs (const char *fake, struct statfs *buf)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_statfs(fake, buf);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_statfs(real, buf);
}

int chmod (const char *fake, mode_t mode)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_chmod(fake, mode);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_chmod(real, mode);
}

int chown (const char *fake, uid_t uid, gid_t gid)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_chown(fake, uid, gid);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_chown(real, uid, gid);
}

int lchown (const char *fake, uid_t uid, gid_t gid)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_lchown(fake, uid, gid);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_lchown(real, uid, gid);
}

int utime (const char *fake, const struct utimbuf *buf)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_utime(fake, buf);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_utime(real, buf);
}

int utimes (const char *fake, const struct timeval *tvp)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_utimes(fake, tvp);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_utimes(real, tvp);
}

int lutimes (const char *fake, const struct timeval *tvp)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_lutimes(fake, tvp);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_lutimes(real, tvp);
}

int link (const char *fake1, const char *fake2)
{
	char real1[PATH_MAX + 1];
	char real2[PATH_MAX + 1];
	if (!mangle) return next_link(fake1, fake2);
	if (fake2real_path(fake1, real1, sizeof(real1)) < 0) return -1;
	if (fake2real_path(fake2, real2, sizeof(real2)) < 0) return -1;
	return next_link(real1, real2);
}

int symlink (const char *fake1, const char *fake2)
{
	char real1[PATH_MAX + 1];
	char real2[PATH_MAX + 1];
	if (!mangle) return next_symlink(fake1, fake2);
	if (fake2real_path(fake1, real1, sizeof(real1)) < 0) return -1;
	if (fake2real_path(fake2, real2, sizeof(real2)) < 0) return -1;
	return next_symlink(real1, real2);
}

int mkdir (const char *fake, mode_t mode)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_mkdir(fake, mode);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_mkdir(real, mode);
}

int mknod (const char *fake, mode_t mode, dev_t dev)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_mknod(fake, mode, dev);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_mknod(real, mode, dev);
}

int rename (const char *fake1, const char *fake2)
{
	char real1[PATH_MAX + 1];
	char real2[PATH_MAX + 1];
	if (!mangle) return next_rename(fake1, fake2);
	if (fake2real_path(fake1, real1, sizeof(real1)) < 0) return -1;
	if (fake2real_path(fake2, real2, sizeof(real1)) < 0) return -1;
	return next_rename(real1, real2);
}

int unlink (const char *fake)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_unlink(fake);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_unlink(real);
}

int rmdir (const char *fake)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_rmdir(fake);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_rmdir(real);
}

int truncate (const char *fake, off_t off)
{
	char real[PATH_MAX + 1];
	if (!mangle) return next_truncate(fake, off);
	if (fake2real_path(fake, real, sizeof(real)) < 0) return -1;
	return next_truncate(real, off);
}

/*--------------------------------------------------------------------------*/
